# -*- coding: utf-8 -*-

from pynfe.xml_sped import *
from lxml import etree
from io import BytesIO
from datetime import datetime, date, time, tzinfo
from decimal import Decimal
import locale
import subprocess
import os
import unicodedata
import re
import pytz

NAMESPACE_NFSE = 'http://homologacaosetelagoas.portalfacil.com.br/nfseserv/schema/nfse_v202.xsd'
NAMESPACE_SIG = 'http://www.w3.org/2000/09/xmldsig#'
ABERTURA = '<?xml version="1.0" encoding="utf-8"?>'


class NohXMLs(object):
    def __init__(self, *args, **kwargs):
        self._xml = None
        self.alertas = []

    def _le_xml(self, arquivo):
        if arquivo is None:
            return False

        if not isinstance(arquivo, str):
            arquivo = etree.tounicode(arquivo)
            #self._xml = arquivo
            #return True

        #elif arquivo is not None:
        if arquivo is not None:
            if isinstance(arquivo, str):
                if '<' in arquivo:
                    self._xml = etree.fromstring(arquivo.replace(ABERTURA, ''))
                else:
                    arq = open(arquivo)
                    txt = ''.join(arq.readlines())
                    txt = tira_abertura(txt)
                    arq.close()
                    self._xml = etree.fromstring(txt)
            else:
                self._xml = etree.parse(arquivo)
            return True

        return False

    def _preenche_namespace(self, tag):
        tag = '/nfe:'.join(tag.split('/')).replace('/nfe:/nfe:', '//nfe:').replace('nfe:sig:', 'sig:')
        return tag

    def _le_nohs(self, tag, ns=None):
        #
        # Tenta ler a tag sem os namespaces
        # Necessário para ler corretamente as tags de grupo reenraizadas
        #
        try:
            nohs = self._xml.xpath(tag)
            if len(nohs) >= 1:
                return nohs
        except:
            pass

        #
        # Não deu certo, tem que botar mesmo os namespaces
        #
        namespaces = {'nfe': NAMESPACE_NFSE, 'sig': NAMESPACE_SIG}

        if ns is not None:
            namespaces['res'] = ns

        if not tag.startswith('//*/res'):
            tag = self._preenche_namespace(tag)

        nohs = self._xml.xpath(tag, namespaces=namespaces)

        if len(nohs) >= 1:
            return nohs
        else:
            return None

    def _le_noh(self, tag, ns=None, ocorrencia=1):
        nohs = self._le_nohs(tag, ns)

        if (nohs is not None) and (len(nohs) >= ocorrencia):
            return nohs[ocorrencia-1]
        else:
            return None

    def _le_tag(self, tag, propriedade=None, ns=None, ocorrencia=1):
        noh = self._le_noh(tag, ns, ocorrencia)

        if noh is None:
            valor = ''
        else:
            if propriedade is None:
                valor = noh.text
            elif (noh.attrib is not None) and (len(noh.attrib) > 0):
                valor = noh.attrib[propriedade]
            else:
                valor = ''

        return valor


class XMLNFSe(NohXMLs):
    def __init__(self, *args, **kwargs):
        super(XMLNFSe, self).__init__(*args, **kwargs)
        self._xml = None
        self.alertas = []
        self.arquivo_esquema = None
        self.caminho_esquema = None

    @property
    def xml(self):
        self.alertas = []
        return ''

    def validar(self):
        arquivo_esquema = self.caminho_esquema + self.arquivo_esquema

        # Aqui eh importante remover a declaracao do `encoding` para evitar erros de conversao
        # `unicode` para ASCII
        xml = self.xml
        xml = xml.replace(ABERTURA, '')
        esquema = etree.XMLSchema(etree.parse(arquivo_esquema))
        if not esquema.validate(etree.fromstring(xml)):
            raise Exception('NFSe nao validou no esquema "%s. \n%s"' % (
                arquivo_esquema, esquema.error_log.last_error))

        return esquema.error_log

    def le_grupo(self, raiz_grupo, classe_grupo, arquivo):
        tags = []

        grupos = self._le_nohs(raiz_grupo)

        if grupos is not None:
            tags = [classe_grupo(raiz=raiz_grupo) for g in grupos]
            for i in range(len(grupos)):
                tags[i].xml = arquivo

        return tags

class ErroObrigatorio(Exception):
    def __init__(self, codigo, nome, propriedade):
        if propriedade:
            self.value = 'No campo código ' + codigo + ', "' + nome + '", a propriedade "' + propriedade + '" é de envio obrigatório, mas não foi preenchida.'
        else:
            self.value = 'O campo código ' + codigo + ', "' + nome + '" é de envio obrigatório, mas não foi preenchido.'

    def __str__(self):
        return repr(self.value)

    def __unicode__(self):
        return str(self.value)

class TamanhoInvalido(Exception):
    def __init__(self, codigo, nome, valor, tam_min=None, tam_max=None, dec_min=None,
                 dec_max=None):
        if tam_min:
            self.value = 'O campo código ' + codigo + ', "' + nome + '", deve ter o tamanho mínimo de ' + str(
                tam_min) + ', mas o tamanho enviado foi ' + str(
                len(str(valor))) + ': ' + str(valor)
        elif tam_max:
            self.value = 'O campo código ' + codigo + ', "' + nome + '", deve ter o tamanho máximo de ' + str(
                tam_max) + ', mas o tamanho enviado foi ' + str(
                len(str(valor))) + ': ' + str(valor)
        elif dec_min:
            self.value = 'O campo código ' + codigo + ', "' + nome + '", deve ter o mínimo de ' + str(
                dec_min) + ' casas decimais, mas o enviado foi ' + str(
                len(str(valor))) + ': ' + str(valor)
        elif dec_max:
            self.value = 'O campo código ' + codigo + ', "' + nome + '", deve ter o máximo de ' + str(
                dec_max) + ' casas decimais, mas o enviado foi ' + str(
                len(str(valor))) + ': ' + str(valor)

    def __str__(self):
        return repr(self.value)

    def __unicode__(self):
        return str(self.value)

class ErroCaracterInvalido(Exception):
    def __init__(self, codigo, nome, propriedade, valor, caracter):
        if propriedade:
            self.value = 'No campo código ' + codigo + ', "' + nome + '", a propriedade "' + propriedade + '" possui um caracter inválido: "' + caracter + '".'
        else:
            self.value = 'O campo código ' + codigo + ', "' + nome + '" possui um caracter inválido: "' + caracter + '".'

    def __str__(self):
        return repr(self.value)

    def __unicode__(self):
        return str(self.value)

class TagCaracter(NohXMLs):
    def __init__(self, *args, **kwargs):
        super(TagCaracter, self).__init__(*args, **kwargs)
        self.codigo = ''
        self.nome = ''
        self._valor_string = ''
        self.obrigatorio = True
        self.tamanho = [None, None, None]
        self.propriedade = None
        self.namespace = None
        self.alertas = []
        self.raiz = None

        # Codigo para dinamizar a criacao de instancias de entidade,
        # aplicando os valores dos atributos na instanciacao
        for k, v in list(kwargs.items()):
            setattr(self, k, v)

        if 'valor' in kwargs:
            self.valor = kwargs['valor']

    def _testa_obrigatorio(self, valor):
        if self.obrigatorio and (not valor):
            return ErroObrigatorio(self.codigo, self.nome, self.propriedade)
            # raise ErroObrigatorio(self.codigo, self.nome, self.propriedade)

    def _testa_tamanho_minimo(self, valor):
        if self.tamanho[0] and (len(str(valor)) < self.tamanho[0]):
            return TamanhoInvalido(self.codigo, self.nome, valor, tam_min=self.tamanho[0])
            # raise TamanhoInvalido(self.codigo, self.nome, valor, tam_min=self.tamanho[0])

    def _testa_tamanho_maximo(self, valor):
        if self.tamanho[1] and (len(str(valor)) > self.tamanho[1]):
            return TamanhoInvalido(self.codigo, self.nome, valor, tam_max=self.tamanho[1])
            # raise TamanhoInvalido(self.codigo, self.nome, valor, tam_max=self.tamanho[1])

    def _valida(self, valor):
        self.alertas = []

        if self._testa_obrigatorio(valor):
            self.alertas.append(self._testa_obrigatorio(valor))

        if self._testa_tamanho_minimo(valor):
            self.alertas.append(self._testa_tamanho_minimo(valor))

        if self._testa_tamanho_maximo(valor):
            self.alertas.append(self._testa_tamanho_maximo(valor))

        return self.alertas == []

    def set_valor(self, novo_valor):
        if novo_valor is not None:
            #
            # Remover caratceres inválidos
            #
            for c in novo_valor:
                if c > 'ÿ':
                    raise ErroCaracterInvalido(self.codigo, self.nome, self.propriedade,
                                               novo_valor, c)

            #
            # É obrigatório remover os espaços no início e no final do valor
            #
            novo_valor = novo_valor.strip()

        if self._valida(novo_valor):
            self._valor_string = novo_valor
        else:
            self._valor_string = ''

    def get_valor(self):
        return self._valor_string

    valor = property(get_valor, set_valor)

    def __unicode__(self):
        if (not self.obrigatorio) and (not self.valor):
            texto = ''
        else:
            texto = '<%s' % self.nome

            if self.namespace:
                texto += ' xmlns="%s"' % self.namespace

            if self.propriedade:
                texto += ' %s="%s">' % (self.propriedade, self._valor_string)
            elif self.valor or (len(self.tamanho) == 3 and self.tamanho[2]):
                texto += '>%s</%s>' % (self._valor_string, self.nome)
            else:
                texto += ' />'

        return texto

    def __repr__(self):
        return self.__unicode__()

    @property
    def xml(self):
        return self.__unicode__()

    def set_xml(self, arquivo, ocorrencia=1):
        if self._le_xml(arquivo):
            self.valor = self._le_tag(self.raiz + '/' + self.nome,
                                      propriedade=self.propriedade, ns=self.namespace,
                                      ocorrencia=ocorrencia)


    def get_text(self):
        if self.propriedade:
            return '%s_%s=%s' % (self.nome, self.propriedade, self._valor_string)
        else:
            return '%s=%s' % (self.nome, self._valor_string)

    text = property(get_text)

    def get_txt(self):
        if self.obrigatorio:
            return self._valor_string

        if self.valor:
            return self._valor_string

        return ''

    txt = property(get_txt)

class TagBoolean(TagCaracter):
    def __init__(self, **kwargs):
        super(TagBoolean, self).__init__(**kwargs)
        self._valor_boolean = None
        # Codigo para dinamizar a criacao de instancias de entidade,
        # aplicando os valores dos atributos na instanciacao
        for k, v in list(kwargs.items()):
            setattr(self, k, v)

        if 'valor' in kwargs:
            self.valor = kwargs['valor']

    def _testa_obrigatorio(self, valor):
        # No caso da tag booleana, False deve ser tratado como preenchido
        if self.obrigatorio and (valor is None):
            return ErroObrigatorio(self.codigo, self.nome, self.propriedade)

    def _valida(self, valor):
        self.alertas = []

        if self._testa_obrigatorio(valor):
            self.alertas.append(self._testa_obrigatorio(valor))

        return self.alertas == []

    def set_valor(self, novo_valor):
        if isinstance(novo_valor, str):
            if novo_valor.lower() == 'true':
                novo_valor = True
            elif novo_valor.lower() == 'false':
                novo_valor = False
            else:
                novo_valor = None

        if isinstance(novo_valor, bool) and self._valida(novo_valor):
            self._valor_boolean = novo_valor

            if novo_valor == None:
                self._valor_string = ''
            elif novo_valor:
                self._valor_string = 'true'
            else:
                self._valor_string = 'false'
        else:
            self._valor_boolean = None
            self._valor_string = ''

    def get_valor(self):
        return self._valor_boolean

    valor = property(get_valor, set_valor)

    def __unicode__(self):
        if (not self.obrigatorio) and (self.valor == None):
            texto = ''
        else:
            texto = '<%s' % self.nome

            if self.namespace:
                texto += ' xmlns="%s"' % self.namespace

            if self.propriedade:
                texto += ' %s="%s">' % (self.propriedade, self._valor_string)
            elif not self.valor == None:
                texto += '>%s</%s>' % (self._valor_string, self.nome)
            else:
                texto += ' />'

        return texto

class TagData(TagCaracter):
    def __init__(self, **kwargs):
        super(TagData, self).__init__(**kwargs)
        self._valor_data = None
        # Codigo para dinamizar a criacao de instancias de entidade,
        # aplicando os valores dos atributos na instanciacao
        for k, v in list(kwargs.items()):
            setattr(self, k, v)

        if 'valor' in kwargs:
            self.valor = kwargs['valor']

    def _valida(self, valor):
        self.alertas = []

        if self._testa_obrigatorio(valor):
            self.alertas.append(self._testa_obrigatorio(valor))

        return self.alertas == []

    def set_valor(self, novo_valor):
        if isinstance(novo_valor, str):
            if novo_valor:
                novo_valor = datetime.strptime(novo_valor, '%Y-%m-%d')
            else:
                novo_valor = None

        if isinstance(novo_valor, (datetime, date)) and self._valida(novo_valor):
            self._valor_data = novo_valor
            # Cuidado!!!
            # Aqui não dá pra usar a função strftime pois em alguns
            # casos a data retornada é 01/01/0001 00:00:00
            # e a função strftime só aceita data com anos a partir de 1900
            self._valor_string = '%04d-%02d-%02d' % (
            self._valor_data.year, self._valor_data.month, self._valor_data.day)
        else:
            self._valor_data = None
            self._valor_string = ''

    def get_valor(self):
        return self._valor_data

    valor = property(get_valor, set_valor)

    def formato_danfe(self):
        if self._valor_data is None:
            return ''
        else:
            return self._valor_data.strftime('%d/%m/%Y')

class TagHora(TagData):
    def set_valor(self, novo_valor):
        if isinstance(novo_valor, str):
            if novo_valor:
                novo_valor = datetime.strptime(novo_valor, '%H:%M:%S')
            else:
                novo_valor = None

        if isinstance(novo_valor, (datetime, time)) and self._valida(novo_valor):
            self._valor_data = novo_valor
            # Cuidado!!!
            # Aqui não dá pra usar a função strftime pois em alguns
            # casos a data retornada é 01/01/0001 00:00:00
            # e a função strftime só aceita data com anos a partir de 1900
            self._valor_string = '%02d:%02d:%02d' % (
            self._valor_data.hour, self._valor_data.minute, self._valor_data.second)
        else:
            self._valor_data = None
            self._valor_string = ''

    def get_valor(self):
        return self._valor_data

    valor = property(get_valor, set_valor)

    def formato_danfe(self):
        if self._valor_data is None:
            return ''
        else:
            return self._valor_data.strftime('%H:%M:%S')

class TagDataHora(TagData):
    def set_valor(self, novo_valor):
        if isinstance(novo_valor, str):
            if novo_valor:
                if not len(novo_valor) > 19:
                    novo_valor = datetime.strptime(novo_valor, '%Y-%m-%dT%H:%M:%S')
            else:
                novo_valor = None

        if isinstance(novo_valor, datetime) and self._valida(novo_valor):
            self._valor_data = novo_valor
            # Cuidado!!!
            # Aqui não dá pra usar a função strftime pois em alguns
            # casos a data retornada é 01/01/0001 00:00:00
            # e a função strftime só aceita data com anos a partir de 1900
            self._valor_string = '%04d-%02d-%02dT%02d:%02d:%02d' % (
            self._valor_data.year, self._valor_data.month, self._valor_data.day,
            self._valor_data.hour, self._valor_data.minute, self._valor_data.second)
        else:
            self._valor_data = None
            self._valor_string = ''

    def get_valor(self):
        return self._valor_data

    valor = property(get_valor, set_valor)

    def formato_danfe(self):
        if self._valor_data is None:
            return ''
        else:
            return self._valor_data.strftime('%d/%m/%Y %H:%M:%S')

class TagDataHoraUTC(TagData):
    def __init__(self, **kwargs):
        super(TagDataHoraUTC, self).__init__(**kwargs)
        #
        # Expressão de validação do formato (vinda do arquivo leiauteSRE_V1.00.xsd
        # Alterada para tornar a informação do fuso horário opcional
        #
        self._validacao = re.compile(
            r'(((20(([02468][048])|([13579][26]))-02-29))|(20[0-9][0-9])-((((0[1-9])|(1[0-2]))-((0[1-9])|(1\d)|(2[0-8])))|((((0[13578])|(1[02]))-31)|(((0[1,3-9])|(1[0-2]))-(29|30)))))T(20|21|22|23|[0-1]\d):[0-5]\d:[0-5]\d(-0[0-4]:00)?')
        self._valida_fuso = re.compile(r'.*-0[0-4]:00$')
        self._brasilia = pytz.timezone('America/Sao_Paulo')
        self.fuso_horario = 'America/Sao_Paulo'

    def set_valor(self, novo_valor):
        if isinstance(novo_valor, str):
            if self._validacao.match(novo_valor):
                if self._valida_fuso.match(novo_valor):
                    #
                    # Extrai e determina qual o fuso horário informado
                    #
                    self.fuso_horario = novo_valor[19:]
                    novo_valor = novo_valor[:19]

                #
                # Converte a data sem fuso horário para o fuso horário atual
                # Isso é necessário pois a função strptime ignora a informação
                # do fuso horário na string de entrada
                #
                novo_valor = self.fuso_horario.localize(
                    datetime.strptime(novo_valor, '%Y-%m-%dT%H:%M:%S'))
            else:
                novo_valor = None

        if isinstance(novo_valor, datetime) and self._valida(novo_valor):
            self._valor_data = novo_valor
            self._valor_data = self._valor_data.replace(microsecond=0)
            try:
                self._valor_data = self.fuso_horario.localize(self._valor_data)
            except:
                pass
            # Cuidado!!!
            # Aqui não dá pra usar a função strftime pois em alguns
            # casos a data retornada é 01/01/0001 00:00:00
            # e a função strftime só aceita data com anos a partir de 1900
            # self._valor_string = '%04d-%02d-%02dT%02d:%02d:%02d' % (self._valor_data.year, self._valor_data.month, self._valor_data.day,
            #    self._valor_data.hour, self._valor_data.minute, self._valor_data.second)

            self._valor_string = self._valor_data.isoformat()
        else:
            self._valor_data = None
            self._valor_string = ''

    def get_valor(self):
        return self._valor_data

    valor = property(get_valor, set_valor)

    def set_fuso_horaro(self, novo_valor):
        if novo_valor in pytz.country_timezones['br']:
            self._fuso_horario = pytz.timezone(novo_valor)

        #
        # Nos valores abaixo, não entendi ainda até agora, mas para o resultado
        # correto é preciso usar GMT+ (mais), não (menos) como seria de se
        # esperar...
        #
        elif novo_valor == '-04:00' or novo_valor == '-0400':
            self._fuso_horario = pytz.timezone('Etc/GMT+4')
        elif novo_valor == '-03:00' or novo_valor == '-0300':
            self._fuso_horario = pytz.timezone('Etc/GMT+3')
        elif novo_valor == '-02:00' or novo_valor == '-0200':
            self._fuso_horario = pytz.timezone('Etc/GMT+2')
        elif novo_valor == '-01:00' or novo_valor == '-0100':
            self._fuso_horario = pytz.timezone('Etc/GMT+1')
        elif novo_valor == '-00:00' or novo_valor == '-0000':
            self._fuso_horario = pytz.timezone('Etc/GMT+0')

    def get_fuso_horario(self):
        return self._fuso_horario

    fuso_horario = property(get_fuso_horario, set_fuso_horaro)

    def formato_danfe(self):
        if self._valor_data is None:
            return ''
        else:
            valor = self._brasilia.normalize(self._valor_data).strftime(
                '%d/%m/%Y %H:%M:%S %Z (%z)')
            #
            # Troca as siglas:
            # BRT - Brazilian Time -> HOB - Horário Oficial do Brasil
            # BRST - Brazilian Summer Time -> HVOB - Horário de Verão Oficial do Brasil
            #
            valor = valor.replace('(-0100)', '(-01:00)')
            valor = valor.replace('(-0200)', '(-02:00)')
            valor = valor.replace('(-0300)', '(-03:00)')
            valor = valor.replace('(-0400)', '(-04:00)')
            valor = valor.replace('BRT', 'HOB')
            valor = valor.replace('BRST', 'HVOB')
            return valor

    def formato_data(self):
        if self._valor_data is None:
            return ''
        else:
            return date.strftime(self._valor_data.date(), "%d-%m-%Y")

    def formato_hora(self):
        if self._valor_data is None:
            return ''
        else:
            return time.strftime(self._valor_data.time(), "%H:%M:%S")

class TagInteiro(TagCaracter):
    def __init__(self, **kwargs):
        super(TagInteiro, self).__init__(**kwargs)
        self._valor_inteiro = 0
        self._valor_string = '0'

        # Codigo para dinamizar a criacao de instancias de entidade,
        # aplicando os valores dos atributos na instanciacao
        for k, v in list(kwargs.items()):
            setattr(self, k, v)

        if 'valor' in kwargs:
            self.valor = kwargs['valor']

    def set_valor(self, novo_valor):
        if isinstance(novo_valor, str):
            if novo_valor:
                novo_valor = int(novo_valor)
            else:
                novo_valor = 0

        if isinstance(novo_valor, (int, Decimal)) and self._valida(novo_valor):
            self._valor_inteiro = novo_valor
            self._valor_string = str(self._valor_inteiro)

            if (len(self.tamanho) >= 3) and self.tamanho[2] and (
                len(self._valor_string) < self.tamanho[2]):
                self._valor_string = self._valor_string.rjust(self.tamanho[2], '0')

        else:
            self._valor_inteiro = 0
            self._valor_string = '0'

    def get_valor(self):
        return self._valor_inteiro

    valor = property(get_valor, set_valor)

    def formato_danfe(self):
        if not (self.obrigatorio or self._valor_inteiro):
            return ''

        return locale.format('%d', self._valor_inteiro, grouping=True)

class TagDecimal(TagCaracter):
    def __init__(self, *args, **kwargs):
        self._valor_decimal = Decimal('0.0')
        self._valor_string = '0.0'
        self.decimais = [None, None, None]
        super(TagDecimal, self).__init__(*args, **kwargs)

        self._valor_decimal = Decimal('0.0')
        self._valor_string = self._formata(self._valor_decimal)
        self.decimais = [None, None, None]

        # Codigo para dinamizar a criacao de instancias de entidade,
        # aplicando os valores dos atributos na instanciacao
        for k, v in list(kwargs.items()):
            setattr(self, k, v)

    def _parte_inteira(self, valor=None):
        if valor is None:
            valor = self._valor_decimal

        valor = str(valor).strip()

        if '.' in valor:
            valor = valor.split('.')[0]

        return valor

    def _parte_decimal(self, valor=None):
        if valor is None:
            valor = self._valor_decimal

        valor = str(valor).strip()

        if '.' in valor:
            valor = valor.split('.')[1]
        else:
            valor = ''

        return valor

    def _formata(self, valor):
        texto = self._parte_inteira(valor)

        dec = self._parte_decimal(valor)
        if not dec:
            dec = '0'

        # Tamanho mínimo das casas decimais
        if (len(self.decimais) >= 3) and self.decimais[2] and (len(dec) < self.decimais[2]):
            dec = dec.ljust(self.decimais[2], '0')

        texto += '.' + dec
        return texto

    def _testa_decimais_minimo(self, decimal):
        if self.decimais[0] and (len(decimal) < self.decimais[0]):
            # return TamanhoInvalido(self.codigo, self.nome, decimal, dec_min=self.decimais[0])
            raise TamanhoInvalido(self.codigo, self.nome, decimal, dec_min=self.decimais[0])

    def _testa_decimais_maximo(self, decimal):
        if self.decimais[1] and (len(decimal) > self.decimais[1]):
            # return TamanhoInvalido(self.codigo, self.nome, decimal, dec_max=self.decimais[1])
            raise TamanhoInvalido(self.codigo, self.nome, decimal, dec_max=self.decimais[1])

    def _valida(self, valor):
        self.alertas = []

        if self._testa_obrigatorio(valor):
            self.alertas.append(self._testa_obrigatorio(valor))

        inteiro = self._parte_inteira(valor)
        decimal = self._parte_decimal(valor)

        if self._testa_tamanho_minimo(inteiro):
            self.alertas.append(self._testa_tamanho_minimo(inteiro))

        if self._testa_tamanho_maximo(inteiro):
            self.alertas.append(self._testa_tamanho_maximo(inteiro))

        if self._testa_decimais_minimo(decimal):
            self.alertas.append(self._testa_decimais_minimo(decimal))

        if self._testa_decimais_maximo(decimal):
            self.alertas.append(self._testa_decimais_maximo(decimal))

        return self.alertas == []

    def set_valor(self, novo_valor):
        if isinstance(novo_valor, str):
            if novo_valor:
                novo_valor = Decimal(novo_valor)
            else:
                novo_valor = Decimal('0.0')

        if isinstance(novo_valor, (int, Decimal)) and self._valida(novo_valor):
            self._valor_decimal = Decimal(novo_valor)
            self._valor_string = self._formata(self._valor_decimal)
        else:
            self._valor_decimal = Decimal('0.0')
            self._valor_string = self._formata(self._valor_decimal)

    def get_valor(self):
        return self._valor_decimal

    valor = property(get_valor, set_valor)

    def formato_danfe(self):
        if not (self.obrigatorio or self._valor_decimal):
            return ''

        # Tamanho mínimo das casas decimais
        if (len(self.decimais) >= 3) and self.decimais[2]:
            if len(self._parte_decimal()) <= self.decimais[2]:
                formato = '%.' + str(self.decimais[2]) + 'f'
            else:
                formato = '%.' + str(len(self._parte_decimal())) + 'f'
        else:
            formato = '%.2f'

        return locale.format(formato, self._valor_decimal, grouping=True)


def formata_datahora(valor, tipo):
    if (tipo == 'd') and isinstance(valor, (datetime, date)):
        valor = '%04d-%02d-%02d' % (valor.year, valor.month, valor.day)
    elif (tipo == 'h') and isinstance(valor, (datetime, time)):
        valor = '%02d:%02d:%02d' % (valor.hour, valor.minute, valor.second)
    elif (tipo == 'dh') and isinstance(valor, datetime):
        valor = '%04d-%02d-%02dT%02d:%02d:%02d' % (
        valor.year, valor.month, valor.day, valor.hour, valor.minute, valor.second)
    return valor


# Tipo de campos do Padrao ABRASF na versao 2.02
# -----------Tipos simples -----------------
class tsNumeroNfse(TagInteiro):
    def __init__(self, **kwargs):
        super(tsNumeroNfse, self).__init__(**kwargs)
        self.tamanho = [15, 15, 1]

class tsCodigoVerificacao(TagCaracter):
    def __init__(self, **kwargs):
        super(tsCodigoVerificacao, self).__init__(**kwargs)
        self.tamanho = [1, 9, 9]

class tsStatusRps(TagInteiro):
    def __init__(self, **kwargs):
        super(tsStatusRps, self).__init__(**kwargs)
        self.tamanho = [1, 1, 1]

    def set_valor(self, novo_valor):
        super(tsStatusRps, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1, 2])


class tsStatusNfse(TagInteiro):
    def __init__(self, **kwargs):
        super(tsStatusNfse, self).__init__(**kwargs)
        self.tamanho = [1, 1, 1]

    def set_valor(self, novo_valor):
        super(tsStatusNfse, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1, 2])

class tsExigibilidadeIss(TagInteiro):
    def __init__(self, **kwargs):
        super(tsExigibilidadeIss, self).__init__(**kwargs)
        self.tamanho = [1, 1, 1]

    def set_valor(self, novo_valor):
        super(tsExigibilidadeIss, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1,2,3,4,5,6,7])

class tsNumeroProcesso(TagCaracter):
    def __init__(self, **kwargs):
        super(tsNumeroProcesso, self).__init__(**kwargs)
        self.tamanho = [1, 30, 30]

class tsRegimeEspecialTributacao(TagInteiro):
    def __init__(self, **kwargs):
        super(tsRegimeEspecialTributacao, self).__init__(**kwargs)
        self.tamanho = [1, 1, 1]

    def set_valor(self, novo_valor):
        super(tsRegimeEspecialTributacao, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1,2,3,4,5,6])

class tsSimNao(TagInteiro):
    def __init__(self, **kwargs):
        super(tsSimNao, self).__init__(**kwargs)
        self.tamanho = [1, 1, 1]

    def set_valor(self, novo_valor):
        super(tsSimNao, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1,2])

class tsResponsavelRetencao(TagInteiro):
    def __init__(self, **kwargs):
        super(tsResponsavelRetencao, self).__init__(**kwargs)
        self.tamanho = [1, 1, 1]

    def set_valor(self, novo_valor):
        super(tsResponsavelRetencao, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1,2])

class tsPagina(TagInteiro):
    def __init__(self, **kwargs):
        super(tsPagina, self).__init__(**kwargs)
        self.tamanho = [1, 6, 1]

class tsNumeroRps(TagInteiro):
    def __init__(self, **kwargs):
        super(tsNumeroRps, self).__init__(**kwargs)
        self.tamanho = [1, 15, 1]

class tsSerieRps(TagCaracter):
    def __init__(self, **kwargs):
        super(tsSerieRps, self).__init__(**kwargs)
        self.tamanho = [1, 5, 5]

class tsTipoRps(TagInteiro):
    def __init__(self, **kwargs):
        super(tsTipoRps, self).__init__(**kwargs)
        self.tamanho = [1, 1, 1]

    def set_valor(self, novo_valor):
        super(tsTipoRps, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1,2,3])


class tsOutrasInformacoes(TagCaracter):
    def __init__(self, **kwargs):
        super(tsOutrasInformacoes, self).__init__(**kwargs)
        self.tamanho = [1, 255, 255]

class tsValor(TagDecimal):
    def __init__(self, **kwargs):
        super(tsValor, self).__init__(**kwargs)
        self.tamanho = [1, 15, 1]
        self.decimais = [0, 2, 2]

class tsItemListaServico(TagCaracter):
    def __init__(self, **kwargs):
        super(tsItemListaServico, self).__init__(**kwargs)
        self.tamanho = [1, 5, 5]

class tsCodigoCnae(TagInteiro):
    def __init__(self, **kwargs):
        super(tsCodigoCnae, self).__init__(**kwargs)
        self.tamanho = [1, 7, 1]

class tsCodigoTributacao(TagCaracter):
    def __init__(self, **kwargs):
        super(tsCodigoTributacao, self).__init__(**kwargs)
        self.tamanho = [1, 20, 20]

class tsAliquota(TagDecimal):
    def __init__(self, **kwargs):
        super(tsAliquota, self).__init__(**kwargs)
        self.tamanho = [1, 4, 1]
        self.decimais = [0, 2, 2]

class tsDiscriminacao(TagCaracter):
    def __init__(self, **kwargs):
        super(tsDiscriminacao, self).__init__(**kwargs)
        self.tamanho = [1, 2000, 2000]

class tsCodigoMunicipioIbge(TagInteiro):
    def __init__(self, **kwargs):
        super(tsCodigoMunicipioIbge, self).__init__(**kwargs)
        self.tamanho = [7, 7, 1]

class tsInscricaoMunicipal(TagCaracter):
    def __init__(self, **kwargs):
        super(tsInscricaoMunicipal, self).__init__(**kwargs)
        self.tamanho = [1, 15, 15]

class tsRazaoSocial(TagCaracter):
    def __init__(self, **kwargs):
        super(tsRazaoSocial, self).__init__(**kwargs)
        self.tamanho = [1, 150, 150]

class tsNomeFantasia(TagCaracter):
    def __init__(self, **kwargs):
        super(tsNomeFantasia, self).__init__(**kwargs)
        self.tamanho = [1, 60, 60]

class tsCnpj(TagCaracter):
    def __init__(self, **kwargs):
        super(tsCnpj, self).__init__(**kwargs)
        self.tamanho = [14, 14, 14]

class tsEndereco(TagCaracter):
    def __init__(self, **kwargs):
        super(tsEndereco, self).__init__(**kwargs)
        self.tamanho = [1, 125, 125]

class tsNumeroEndereco(TagCaracter):
    def __init__(self, **kwargs):
        super(tsNumeroEndereco, self).__init__(**kwargs)
        self.tamanho = [1, 10, 10]

class tsComplementoEndereco(TagCaracter):
    def __init__(self, **kwargs):
        super(tsComplementoEndereco, self).__init__(**kwargs)
        self.tamanho = [1, 60, 60]

class tsBairro(TagCaracter):
    def __init__(self, **kwargs):
        super(tsBairro, self).__init__(**kwargs)
        self.tamanho = [1, 60, 60]

class tsUf(TagCaracter):
    def __init__(self, **kwargs):
        super(tsUf, self).__init__(**kwargs)
        self.tamanho = [2, 2, 2]

class tsCodigoPaisBacen(TagCaracter):
    def __init__(self, **kwargs):
        super(tsCodigoPaisBacen, self).__init__(**kwargs)
        self.tamanho = [4, 4, 4]

class tsCep(TagCaracter):
    def __init__(self, **kwargs):
        super(tsCep, self).__init__(**kwargs)
        self.tamanho = [8, 8, 8]

class tsEmail(TagCaracter):
    def __init__(self, **kwargs):
        super(tsEmail, self).__init__(**kwargs)
        self.tamanho = [1, 80, 80]

class tsTelefone(TagCaracter):
    def __init__(self, **kwargs):
        super(tsTelefone, self).__init__(**kwargs)
        self.tamanho = [1, 20, 20]

class tsCpf(TagCaracter):
    def __init__(self, **kwargs):
        super(tsCpf, self).__init__(**kwargs)
        self.tamanho = [11, 11, 11]

class tsCodigoObra(TagCaracter):
    def __init__(self, **kwargs):
        super(tsCodigoObra, self).__init__(**kwargs)
        self.tamanho = [1, 15, 15]

class tsArt(TagCaracter):
    def __init__(self, **kwargs):
        super(tsArt, self).__init__(**kwargs)
        self.tamanho = [1, 15, 15]

class tsNumeroLote(TagInteiro):
    def __init__(self, **kwargs):
        super(tsNumeroLote, self).__init__(**kwargs)
        self.tamanho = [15, 15, 1]

class tsNumeroProtocolo(TagCaracter):
    def __init__(self, **kwargs):
        super(tsNumeroProtocolo, self).__init__(**kwargs)
        self.tamanho = [0, 50, 50]

class tsNumeroLote(TagInteiro):
    def __init__(self, **kwargs):
        super(tsNumeroLote, self).__init__(**kwargs)
        self.tamanho = [1, 15, 1]

    def set_valor(self, novo_valor):
        super(tsNumeroLote, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1, 2, 3, 4])

class tsCodigoMensagemAlerta(TagCaracter):
    def __init__(self, **kwargs):
        super(tsCodigoMensagemAlerta, self).__init__(**kwargs)
        self.tamanho = [1, 4, 4]

class tsDescricaoMensagemAlerta(TagCaracter):
    def __init__(self, **kwargs):
        super(tsDescricaoMensagemAlerta, self).__init__(**kwargs)
        self.tamanho = [1, 200, 200]

class tsSituacaoLoteRps(TagInteiro):
    def __init__(self, **kwargs):
        super(tsSituacaoLoteRps, self).__init__(**kwargs)
        self.tamanho = [1, 1, 1]

    def set_valor(self, novo_valor):
        super(tsSituacaoLoteRps, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1, 2, 3, 4])

class tsQuantidadeRps(TagInteiro):
    def __init__(self, **kwargs):
        super(tsQuantidadeRps, self).__init__(**kwargs)
        self.tamanho = [1, 4, 1]

class tsCodigoCancelamentoNfse(TagInteiro):
    def __init__(self, **kwargs):
        super(tsCodigoCancelamentoNfse, self).__init__(**kwargs)
        self.tamanho = [1, 4, 1]

    def set_valor(self, novo_valor):
        super(tsCodigoCancelamentoNfse, self).set_valor(novo_valor)
        _valida_valor_inteiro(self, novo_valor, [1, 2, 3, 4, 5])

def _valida_valor_inteiro(self, novo_valor, valores_validos=[]):
    ''' Valida se o valor a ser inserido na tag esta de acordo com os valores possiveis
    :param self: 
    :param novo_valor: valor a ser inserido na tag
    :param valores_validos: Lista com os valores validos
    :return: 
    '''
    if novo_valor not in valores_validos:
        raise Exception("Valor do campo '%s' invalido, valores aceitaveis: '%s'\n Valor escolhido: %s" %
                        (self.name, ', '.join(['%s' % x for x in valores_validos]),
                         novo_valor))
